// Find out what happens when a packet larger than the interface MTU is written
// to a socket - but first you have to guess what will happen, and it will
// output a score at the end.
//
// Copyright (c) Valerie Aurora https://valerieaurora.org/

package main

import (
	"errors"
	"flag"
	"fmt"
	"log"
	"net"
	"net/netip"
	"os"
	"runtime"
	"strconv"
	"strings"
	"syscall"
	"time"

	"golang.org/x/net/icmp"
	"golang.org/x/net/ipv4"
	"golang.org/x/net/ipv6"
)

type fragQuiz struct {
	// ICMP sockets for receiving ICMP TTL exceeded messages
	// ICMP version string -> icmp.PacketConn
	icmp map[string]icmp.PacketConn
	// Trials to run
	trials []*trial
	// Score
	correct, total int
}

// Command line options
var verbose bool
var noGuess bool
var randomGuess bool
var noPause bool
var remoteHostPortV4 string
var remoteHostPortV6 string

// XXX should be different for v4 and v6
var defaultInterface string
var readTimeoutInt int
var dialTimeoutInt int
var testProtosStr string
var packetSizesStr string
var fragValuesStr string
var ttl4 int
var ttl6 int
var retries int

// Things derived from command line options

var readTimeout time.Duration
var dialTimeout time.Duration

// List of protocols to test with associated address to dial
// proto name -> host:port
var protoAddrPorts map[string]string

// List of packet sizes to test
var packetSizes []int

// List of don't fragment values to test
var dontFragValues []int

// Default don't fragment value on a socket on this system (diff from system setting in Linux)
var defaultDontFragSetting int

// It is important to differentiate between different "kinds" of fragmentation
// "bits":
//
// 1. The Don't Fragment bit in IPv4 packet headers. This is set when the packet
// is produced based on a number of inputs.
//
// 2. The More Fragments header in IPv6 packet headers. This is set when the
// packet is fragmented at the source.
//
// 3. The system-wide default setting for path MTU discovery. This is an
// indirect influence on the fragmentation status of the eventual output packet.
//
// 4. The DONTFRAG socket option on this specific socket. This defaults to a
// certain value based on the system-wide path MTU discovery setting but can be
// set by the user. This can be set on both IPv4 and IPv6 sockets, but IPv6 only
// ever produces either a fragmented packet or EMSGSIZE, so you can't ask "what
// will the don't frag bit be?" on an IPv6 packet.

type result struct {
	noResult   bool
	tooBig     string
	dfBit      string
	fragmented string
	length     int
}

type trial struct {
	network         string
	size            int
	dontFragSetting int
	ttl             int
	retries         int
	host            string
	guess           result
	answer          result
}

func init() {
	flag.BoolVar(&verbose, "v", false, "print verbose output")
	flag.BoolVar(&noGuess, "n", false, "don't ask for a guess")
	flag.BoolVar(&randomGuess, "r", false, "fill in a random guess")
	flag.BoolVar(&noPause, "c", false, "don't pause after each test (continuous mode)")
	flag.StringVar(&remoteHostPortV4, "a", "google.com:80", "hostname:port to connect for IPv4 sockets")
	flag.StringVar(&remoteHostPortV6, "A", "google.com:80", "hostname:port to connect for IPv6 sockets")
	flag.StringVar(&defaultInterface, "i", "en0", "default interface to use for IPv6 tests")
	flag.IntVar(&readTimeoutInt, "t", 500, "number of milliseconds to wait for ICMP reads")
	flag.IntVar(&dialTimeoutInt, "d", 500, "number of milliseconds to wait to connect to servers")
	flag.StringVar(&testProtosStr, "p", "udp4,tcp4,udp6,tcp6", "protocols to test, comma-separated (udp4,tcp4,udp6,tcp6)")
	flag.StringVar(&packetSizesStr, "s", "2000", "packet sizes to test (between 48 and 65536, comma-separated)")
	flag.IntVar(&ttl4, "l", 0, "time-to-live or hop limit value to use for IPv4 probe packets (0 to probe for best TTL)")
	flag.IntVar(&ttl6, "L", 0, "time-to-live or hop limit value to use for IPv4 probe packets (0 to probe for best TTL)")
	flag.IntVar(&retries, "R", 2, "number of times to retry any test if we get no result")

	var fragHelp string
	switch runtime.GOOS {
	case "darwin":
		fragHelp = "fragmentation settings to test (default,dontfrag,dofrag)"
	case "linux":
		fragHelp = "fragmentation settings to test (default,dontfrag,dofrag,wantpmtud,probe)"
	default:
		log.Fatalf("unsupported OS %v\n", runtime.GOOS)
	}
	flag.StringVar(&fragValuesStr, "f", "default,dontfrag", fragHelp)
}

func main() {
	log.SetFlags(log.Lshortfile)
	processArgs()

	f := initSockets()

	f.generateTrials()

	f.printSystemInfo()

	f.runTrials()

	f.printScoreSummary()
}

func processArgs() {
	flag.Parse()

	readTimeout = time.Duration(readTimeoutInt) * time.Millisecond
	dialTimeout = time.Duration(dialTimeoutInt) * time.Millisecond

	// Parse the protocols to test string
	protoAddrPorts = make(map[string]string)
	for _, p := range strings.Split(testProtosStr, ",") {
		switch p {
		case "tcp4", "udp4":
			protoAddrPorts[p] = remoteHostPortV4
		case "tcp6", "udp6":
			protoAddrPorts[p] = remoteHostPortV6
		default:
			log.Fatalf("unknown protocol: %v\n", p)
		}
	}

	if len(protoAddrPorts) == 0 {
		log.Fatalf("no protocols specified in string: %v\n", testProtosStr)
	}

	// Parse the packet sizes to test string
	sizes := strings.Split(packetSizesStr, ",")
	for _, str := range sizes {
		s, err := strconv.Atoi(str)
		if err != nil {
			log.Fatalf("Bad packet size specification %s: %v\n", str, err)
		}
		// XXX max should be default interface MTU
		if s < 48 || s > 65536 {
			log.Fatalf("Invalid packet size %s: %v (must be between 48 and 65536)\n", str, err)
		}
		packetSizes = append(packetSizes, s)
	}

	// We parse the fragmentation settings after we know the default setting,
	// which we have to open the sockets first to find out.
}

// For each IP family, find and resolve a host address/port combo that we can
// connect a socket to for tests.
func (f *fragQuiz) findHostsForFamily(family string) {
	var tcpNetwork, udpNetwork string
	switch family {
	default:
		log.Fatalf("unknown family %v\n", family)
	case "ipv4":
		tcpNetwork = "tcp4"
		udpNetwork = "udp4"
	case "ipv6":
		tcpNetwork = "tcp6"
		udpNetwork = "udp6"
	}
	// Look for TCP first
	f.findTCPHost(tcpNetwork)

	// Now connect UDP, piggybacking off of TCP's success, if any
	f.findUDPHost(udpNetwork)
}

// Find addresses to connect the sockets to. In order of preference:
//
// Actual real server somewhere on the internet
// Random node with an open TCP port on the local link
//
// Success or failure, it updates the protoAddrPorts list.
func (f *fragQuiz) findTCPHost(network string) {
	addrPort, ok := protoAddrPorts[network]
	if !ok {
		return
	}

	if verbose {
		fmt.Printf("Connecting to %v...\n", addrPort)
	}

	// Simplest case: we have a functioning route to a remote server.
	c, err := net.DialTimeout(network, addrPort, dialTimeout)
	if err == nil {
		if verbose {
			fmt.Printf("Using %v for %v tests\n", c.RemoteAddr(), network)
		}
		c.Close()
		return
	}
	// Failure. This host won't do.
	delete(protoAddrPorts, network)
	fmt.Printf("Couldn't connect to specified host %v for %v tests: %v\n", addrPort, network, err)

	return

	// XXX Link local IPv6 doesn't work yet - it needs to force routing through the
	// router, and if that router reassembles packets before sending hop limit
	// exceeded, it won't work anyway. So skip it.
	//
	// Okay, no functioning connection to the outside world. See if there's
	// another device on the local link that will let us connect.
	//err = f.scanTCPServer(network)
	//if err != nil {
	//		fmt.Printf("Scan for %v server failed: %v\n", network, err)
	//}
}

// Find a UDP host we can "connect" to for this network.
func (f *fragQuiz) findUDPHost(network string) error {
	udpAddrPort, ok := protoAddrPorts[network]
	if !ok {
		fmt.Printf("skipping %v\n", network)
		// User did not request testing this protocol
		return nil
	}

	c, err := net.DialTimeout(network, udpAddrPort, dialTimeout)
	if err == nil {
		if verbose {
			fmt.Printf("Using %v for %v tests\n", c.RemoteAddr(), network)
		}
		c.Close()
		return nil
	}
	// Boo. Delete the non-working entry
	delete(protoAddrPorts, network)

	// Try the matching TCP addr/port
	var tcpAddrPort string
	switch network {
	case "udp4":
		tcpAddrPort = protoAddrPorts["tcp4"]
	case "udp6":
		tcpAddrPort = protoAddrPorts["tcp6"]
	default:
		log.Fatalf("unknown network %v\n", network)
	}
	c, err = net.DialTimeout(network, tcpAddrPort, dialTimeout)
	if err == nil {
		// Hurray! Update the addr/port entry
		protoAddrPorts[network] = tcpAddrPort
		if verbose {
			fmt.Printf("Using %v for %v tests\n", c.RemoteAddr(), network)
		}
		c.Close()
	}
	return err
}

// Get a raw conn so we can set socket options
func getRawSocket(c net.Conn) syscall.RawConn {
	var r syscall.RawConn
	var err error
	switch t := c.(type) {
	case *net.TCPConn:
		r, err = t.SyscallConn()
	case *net.UDPConn:
		r, err = t.SyscallConn()
	default:
		log.Panicf("unknown conn type %v\n", t)
	}
	if err != nil {
		log.Fatalf("couldn't get raw conn for socket: %v\n", err)
	}
	return r
}

func initSockets() *fragQuiz {
	f := &fragQuiz{
		icmp: make(map[string]icmp.PacketConn),
	}

	var anySuccess bool
	// Start the ICMP listeners
	err := f.initICMP("icmp4", "0.0.0.0")
	if errors.Is(err, os.ErrPermission) {
		log.Fatal("Sorry, on this OS you need superuser privileges to listen to ICMP TTL exceeded messages. Switch to MacOS or rerun with sudo.\n")
	}
	if err == nil {
		anySuccess = true
	}
	err = f.initICMP("icmp6", "::")
	if err == nil {
		anySuccess = true
	}

	if !anySuccess {
		log.Fatal("Couldn't open any ICMP sockets! Check your configuration.")
	}

	// Figure out what address/ports to use for each protocol to test
	f.findHostsForFamily("ipv4")
	f.findHostsForFamily("ipv6")

	// Did we open any socket successfully?
	if len(protoAddrPorts) == 0 {
		log.Fatal("Couldn't find hosts to test any of the protocols!\n")
	}

	// Now that we have can open sockets to find the default fragmentation
	// setting from, parse the fragmentation options string.
	for network, addrPort := range protoAddrPorts {
		c, err := net.DialTimeout(network, addrPort, dialTimeout)
		if err != nil {
			log.Fatalf("Couldn't open a socket for %v network\n", network)
		}
		r := getRawSocket(c)
		defaultDontFragSetting = getFrag(network, r)
		c.Close()
		break
	}
	parseFrags()

	// Probe to find out how many hops we have to go before we get a router that
	// is not reassembling packets before sending TTL exceeded.
	// XXX only do this for sockets we are using
	f.probeHops(remoteHostPortV4, "IPv4", "udp4", &ttl4)
	f.probeHops(remoteHostPortV6, "IPv6", "udp6", &ttl6)
	return f
}

func (f *fragQuiz) probeHops(addrPort, family, network string, ttl *int) {
	// Are we testing in this family?
	var host string
	switch family {
	case "IPv4":
		a1 := protoAddrPorts["udp4"]
		a2 := protoAddrPorts["tcp4"]
		if a1 == "" && a2 == "" {
			return
		}
		if a1 == "" {
			host = a2
		} else {
			host = a1
		}
	case "IPv6":
		a1 := protoAddrPorts["udp6"]
		a2 := protoAddrPorts["tcp6"]
		if a1 == "" && a2 == "" {
			return
		}
		if a1 == "" {
			host = a2
		} else {
			host = a1
		}
	default:
		log.Panicf("unknown family %s\n", family)
	}

	if *ttl != 0 {
		fmt.Printf("Using user-specified TTL %v for %v tests\n", *ttl, family)
		return
	}
	fmt.Printf("Probing for a TTL that gets replies that haven't been reassembled\n")
	fmt.Printf("To prevent this probe, use the -l and -L options to set the TTL manually\n")
	for hops := 1; hops < 10; hops++ {
		t := trial{network: network, size: 2000, dontFragSetting: getDoFragValue(), ttl: hops, retries: 0, host: host}
		f.runOneTrial(&t)
		// Check the fragmentation status of the trial
		if t.answer.fragmented == "yes" {
			*ttl = hops
			fmt.Printf("Using probed TTL %v for %v tests\n", *ttl, family)
			return
		}
	}
	fmt.Printf("Couldn't find a TTL that gets replies from a router that does not reassemble %v packets, skipping\n", family)
	switch family {
	case "IPv4":
		delete(protoAddrPorts, "udp4")
		delete(protoAddrPorts, "tcp4")
	case "IPv6":
		delete(protoAddrPorts, "udp6")
		delete(protoAddrPorts, "tcp6")
	default:
		log.Panicf("unknown family %s\n", family)
	}

}

func (f *fragQuiz) generateTrials() {
	// Build our trial structs
	for n := range protoAddrPorts {
		var ttl int
		switch n {
		case "udp4", "tcp4":
			ttl = ttl4
		case "udp6", "tcp6":
			ttl = ttl6
		default:
			log.Fatalf("unknown network %v\n", n)
		}
		for _, s := range packetSizes {
			for _, v := range dontFragValues {
				if verbose {
					fmt.Printf("Creating trial network %v size %v dontFrag %v\n", n, s, v)
				}
				f.trials = append(f.trials, &trial{network: n, size: s, dontFragSetting: v, ttl: ttl, retries: retries, host: protoAddrPorts[n]})
			}
		}
	}
}

func (f *fragQuiz) printSystemInfo() {
	fmt.Printf("This is a %v system\n", runtime.GOOS)
	fmt.Printf("The system-wide Path MTU Discovery (PMTUD) setting is %v\n", printPMTUD())
	fmt.Printf("The don't fragment (IP_DONTFRAG) socket option is set to %v by default\n", printFragSetting(defaultDontFragSetting))
}

func (t *trial) collectTrialGuess() {
	if noGuess {
		return
	}
	r := &t.guess
	var askDF, askTooBig bool
	switch t.network {
	case "udp4":
		askTooBig = true
		askDF = true
	case "tcp4":
		askDF = true
	case "udp6":
		askTooBig = true
	default:
		// Don't need to do anything for tcp6
	}

	fmt.Printf("If we send a packet %v bytes long on a %v socket with %v socket option:\n", t.size, t.network, printFragSetting(t.dontFragSetting))

	if askTooBig {
		r.tooBig = collectOneGuess("Will it be rejected because it is too big?")
	}

	// If the user guesses that the packet gets rejected for being too big, then
	// it must have had don't fragment set and was not fragmented. Count that as
	// one guess/answer, not three.
	if r.tooBig != "yes" {
		if askDF {
			r.dfBit = collectOneGuess("Will the don't fragment bit be set?")
		}
		// Likewise, if the DF bit is set, then we know it can't be fragmented,
		// so don't ask about fragmentation separately.
		if r.dfBit != "yes" {
			r.fragmented = collectOneGuess("Will the packet be fragmented?")
		}
	}
}

var lastGuess = "no"

func collectOneGuess(desc string) string {
	fmt.Printf("%v ", desc)
	if randomGuess {
		// Simple alternating "random" guess
		if lastGuess == "yes" {
			lastGuess = "no"
		} else {
			lastGuess = "yes"
		}
		fmt.Printf("%v\n", lastGuess)
		return lastGuess
	}

	for {
		var g string
		fmt.Scanln(&g)
		switch g {
		case "n", "N", "no":
			return "no"
		case "y", "Y", "yes":
			return "yes"
		case "q", "Q", "quit":
			os.Exit(0)
		default:
			fmt.Printf("Unrecognized input %v, please use n/N/no/y/Y/yes or q/Q/quit to quit\n", g)
		}
	}
}

// Send a test packet with the requested parameters. Opens and closes a new
// socket every time to avoid previous tests affecting the current test.
// XXX: Need to clear OS path MTU cache, or at least have that be a parameter.
func (f *fragQuiz) sendPacket(t *trial) error {
	var payloadLen int
	switch t.network {
	case "udp4", "tcp4":
		payloadLen = t.size - 28
	case "udp6", "tcp6":
		payloadLen = t.size - 48
	default:
		log.Panicf("unknown network %v\n", t.network)
	}

	switch t.network {
	case "tcp4", "tcp6":
		fmt.Printf("Testing: %v packet length %v bytes %v ttl\n", t.network, t.size, t.ttl)
	default:
		fmt.Printf("Testing: %v packet length %v bytes with %v socket option ttl %v\n", t.network, t.size, printFragSetting(t.dontFragSetting), t.ttl)
	}

	c, err := net.DialTimeout(t.network, t.host, dialTimeout)
	if err != nil {
		return err
	}
	r := getRawSocket(c)
	// Note that darwin only allows setting the fragmentation option once for a
	// TCP socket, which is fine because we recreate the socket for each trial.
	setFrag(t.network, r, t.dontFragSetting)

	// Set the TTL so we get a TTL exceeded packet in reply
	// XXX add some auto-probe code to get a useful value
	f.setTTL(t.network, r, t.ttl)

	b := make([]byte, payloadLen)
	copy([]byte("I'm playing fragquiz https://valerieaurora.org/fragquiz.html"), b)

	_, err = c.Write(b)
	c.Close()
	return err
}

func (f *fragQuiz) sendICMP(network string, addr string, payloadLen int) {
	if verbose {
		fmt.Printf("Sending %v echo packet to %v payload length %v\n", network, addr, payloadLen)
	}
	var t icmp.Type
	var c icmp.PacketConn
	var ok bool
	switch network {
	default:
		log.Fatalf("unknown network %v\n", network)
	case "udp4", "tcp4", "icmp4":
		t = ipv4.ICMPTypeEcho
		c, ok = f.icmp["icmp4"]
	case "udp6", "tcp6", "icmp6":
		t = ipv6.ICMPTypeEchoRequest
		c, ok = f.icmp["icmp6"]
	}
	if !ok {
		log.Fatalf("no icmp socket for %v\n", network)
	}

	ip, err := netip.ParseAddr(addr)
	if err != nil {
		log.Fatalf("can't parse addr %v\n", addr)
	}
	wm := icmp.Message{
		Type: t, Code: 0,
		Body: &icmp.Echo{
			ID: os.Getpid() & 0xffff, Seq: 1,
			Data: make([]byte, payloadLen),
		},
	}
	wb, err := wm.Marshal(nil)
	if err != nil {
		log.Fatal(err)
	}

	_, err = (&c).WriteTo(wb, &net.UDPAddr{IP: ip.AsSlice(), Zone: ip.Zone()})
	if err != nil {
		log.Fatal(err)
	}
}

// Look for an ICMP TTL response to our packet and if we find one, record
// it. Then read and throw away any remaining ICMP messages so we can start the
// next trial.
func (f *fragQuiz) readICMPTTL(t *trial) error {
	var err error
	var reply *icmp.Message
	for {
		reply, _, err = f.readICMP(t.network)
		if err != nil {
			break
		}
		switch t.network {
		case "udp4", "tcp4":
			if reply.Type != ipv4.ICMPTypeTimeExceeded {
				continue
			}
		case "udp6", "tcp6":
			if reply.Type != ipv6.ICMPTypeTimeExceeded {
				continue
			}
		}
		t.processICMPReply(reply)
		break
	}
	if err != nil {
		t.answer.noResult = true
		fmt.Printf("No result: did not get an ICMP message\n")
		return errors.New("no result")
	}

	// Sometimes you get a LOT of ICMP messages in response to one packet. Drain
	// the queue before starting the next trial.
	for err == nil {
		reply, _, err = f.readICMP(t.network)
		if err == nil && verbose {
			fmt.Printf("readICMPTTL: discarding ICMP packet %v\n", reply.Type)
		}
	}
	return nil
}

func (f *fragQuiz) readICMP(network string) (*icmp.Message, string, error) {
	var i icmp.PacketConn
	var proto int
	switch network {
	case "udp4", "tcp4", "icmp4":
		i = f.icmp["icmp4"]
		// seriously? surely there is a def?
		proto = 1
	case "udp6", "tcp6", "icmp6":
		i = f.icmp["icmp6"]
		proto = 58
	default:
		log.Panicf("unknown network %v\n", network)
	}

	rb := make([]byte, 65536)
	i.SetReadDeadline(time.Now().Add(readTimeout))
	n, srcAddr, err := i.ReadFrom(rb)
	if err != nil {
		if !errors.Is(err, os.ErrDeadlineExceeded) {
			fmt.Printf("ICMP read failed: %v\n", err)
		} else if verbose {
			fmt.Printf("ICMP read returned: %v\n", err)
		}
		return nil, "", err
	}

	reply, err := icmp.ParseMessage(proto, rb[:n])
	if err != nil {
		log.Fatal(err)
	}
	srcIP, _, _ := net.SplitHostPort(srcAddr.String())
	if verbose {
		fmt.Printf("Read %v packet type %v src %v\n", network, reply.Type, srcIP)
	}
	return reply, srcIP, nil
}

// Parse the ICMP message we got in reply to our packet and print theh result.
func (t *trial) processICMPReply(reply *icmp.Message) {
	if verbose {
		fmt.Printf("Processing ICMP reply type %v\n", reply.Type)
	}
	switch reply.Type {
	case ipv4.ICMPTypeTimeExceeded:
		body, ok := reply.Body.(*icmp.TimeExceeded)
		if !ok {
			log.Fatal("couldn't parse ICMP reply")
		}
		// The ICMP TTL exceeded body contains the IP header of the datagram
		// that prompted the reply. So look at the IP header to see if it shows
		// that it was fragmented.
		sentIPHeader, err := icmp.ParseIPv4Header(body.Data)
		if err != nil {
			log.Fatal(err)
		}
		if sentIPHeader.Flags == ipv4.DontFragment {
			t.answer.dfBit = "yes"
		} else {
			t.answer.dfBit = "no"
		}
		if t.answer.dfBit != "yes" {
			if sentIPHeader.Flags == ipv4.MoreFragments {
				t.answer.fragmented = "yes"
			} else {
				t.answer.fragmented = "no"
			}
		}
		t.answer.length = sentIPHeader.TotalLen
	case ipv6.ICMPTypeTimeExceeded:
		body, ok := reply.Body.(*icmp.TimeExceeded)
		if !ok {
			log.Fatal("couldn't parse ICMP reply")
		}
		// The ICMP TTL exceeded body contains the IP header of the datagram
		// that prompted the reply. So look at the IP header to see if it shows
		// that it was fragmented.
		sentIPHeader, err := ipv6.ParseHeader(body.Data)
		if err != nil {
			log.Fatal(err)
		}
		// In IPv6, a fragmented packet is identified by the present of a
		// Fragmentation Header. Scan through the headers to see if there is a
		// fragmentation header.
		if sentIPHeader.NextHeader == 44 {
			t.answer.fragmented = "yes"
		} else {
			t.answer.fragmented = "no"
		}
		t.answer.length = sentIPHeader.PayloadLen + ipv6.HeaderLen
	default:
		fmt.Printf("No result: received ICMP packet type %v instead of TTL exceeded\n", reply.Type)
	}
}

func comparePackets() {}

func (f *fragQuiz) runTrials() {
	for _, t := range f.trials {
		t.collectTrialGuess()
		f.runOneTrial(t)
		f.printOneScore(t)

		if !noPause {
			fmt.Printf("Hit enter to run the next test\n")
			var tmp string
			fmt.Scanln(&tmp)
		}
	}
}

func (f *fragQuiz) runOneTrial(t *trial) {
	for i := 0; i < t.retries+1; i++ {
		err := f.sendPacket(t)
		if err != nil {
			if errors.Is(err, syscall.EMSGSIZE) {
				t.answer.tooBig = "yes"
				break
			} else {
				fmt.Printf("sendPacket: unexpected error: %v\n", err)
			}
		} else {
			switch t.network {
			case "udp4", "udp6":
				t.answer.tooBig = "no"
			}
			err := f.readICMPTTL(t)
			if err == nil {
				break
			}
		}
		fmt.Printf("Didn't get an ICMP TTL exceeded message, retrying %v more times\n", t.retries-i)
	}
}

func (f *fragQuiz) printOneScore(t *trial) {
	if t.answer.noResult {
		fmt.Printf("No result, can't score this test :(\n")
		return
	}

	f.scoreOne(t.guess.tooBig, t.answer.tooBig, "packet too big", t)
	f.scoreOne(t.guess.dfBit, t.answer.dfBit, "don't fragment bit", t)
	f.scoreOne(t.guess.fragmented, t.answer.fragmented, "packet fragmented", t)
	fmt.Printf(t.answer.formatResult("Result"))
	fmt.Printf(t.guess.formatResult("Guess"))
	fmt.Printf("Your score so far: %v/%v\n", f.correct, f.total)
}

func (f *fragQuiz) scoreOne(guess, answer, desc string, t *trial) {
	// If we didn't get a result, don't score it
	if t.answer.noResult {
		fmt.Printf("Skipping scoring of %v for %v packet size %v df %v because it could not be tested\n",
			desc, t.network, t.size, t.dontFragSetting)
		return
	}

	// Both guess and answer are ""? then don't score this
	if guess == "" && answer == "" {
		return
	}

	f.total++
	if guess == answer {
		f.correct++
	}

	if verbose {
		fmt.Printf("Current score %v/%v (%v guess %v answer %v)\n", f.correct, f.total, desc, guess, answer)
	}
}

func (r *result) formatResult(label string) string {
	// Assemble the output
	scoreStr := fmt.Sprintf("%s:\tpacket too big: %v", label, r.tooBig)
	if r.tooBig != "yes" {
		if r.dfBit != "" {
			scoreStr = scoreStr + " don't fragment bit: " + r.dfBit
		}
		if r.dfBit != "yes" {
			scoreStr = scoreStr + " fragmented: " + r.fragmented
		}
	}
	if r.length != 0 {
		scoreStr = scoreStr + " (length: " + strconv.Itoa(r.length) + ")"
	}
	scoreStr = scoreStr + "\n"
	return scoreStr
}

func (f *fragQuiz) printScoreSummary() {
	// XXX Make this involve emoji or something

	fmt.Printf("\n")
	fmt.Printf("Summary:\n")
	fmt.Printf("\n")
	fmt.Printf("%-6s %-6s %-30s | %6s %6s %s\n",
		"proto", "size", "DF setting", "too big", "DF bit", "fragmented")
	fmt.Printf("------------------------------------------------------------------------\n")
	for _, t := range f.trials {
		fmt.Printf("%-6s %-6d %-30s | ", t.network, t.size, printFragSetting(t.dontFragSetting))
		fmt.Printf(" %-6s %-6s %s\n",
			t.getOneWordScore(t.guess.tooBig, t.answer.tooBig),
			t.getOneWordScore(t.guess.dfBit, t.answer.dfBit),
			t.getOneWordScore(t.guess.fragmented, t.answer.fragmented))
	}
	fmt.Printf("\n")
	fmt.Printf("Thank you for playing fragquiz! Share this with your friends:\n")
	fmt.Printf("\n")
	for _, t := range f.trials {
		fmt.Printf("%s%s%s\n",
			t.getOneCharScore(t.guess.tooBig, t.answer.tooBig),
			t.getOneCharScore(t.guess.dfBit, t.answer.dfBit),
			t.getOneCharScore(t.guess.fragmented, t.answer.fragmented))
	}
	fmt.Printf("My #fragquiz score: %v/%v\n", f.correct, f.total)
	fmt.Printf("https://valerieaurora.org/fragquiz.html\n")
}

func (t *trial) getOneWordScore(guess, answer string) string {
	if t.answer.noResult {
		return "?"
	}
	if answer != guess {
		return "wrong"
	}
	if answer == "" {
		return "N/A"
	}
	return "right"
}

func (t *trial) getOneCharScore(guess, answer string) string {
	s := t.getOneWordScore(guess, answer)
	switch s {
	case "wrong":
		return "-"
	case "right":
		return "+"
	case "N/A":
		return "."
	case "?":
		return "?"
	default:
		log.Fatalf("unknown score %v\n", s)
	}
	// Unreached
	return ""
}
