// Copyright (c) Valerie Aurora https://valerieaurora.org/
//
// See COPYING for license information

package main

import (
	"fmt"
	"log"
	"strings"
	"syscall"

	"golang.org/x/net/icmp"
	"golang.org/x/sys/unix"
)

// Parse the frag values to test string
func parseFrags() {
	frags := strings.Split(fragValuesStr, ",")
	for _, str := range frags {
		dontFragValues = append(dontFragValues, parseFrag(str))
	}
}

func parseFrag(str string) int {
	switch str {
	case "default":
		return defaultDontFragSetting
	case "dofrag":
		return 0
	case "dontfrag":
		return 1
	default:
		log.Fatalf("Bad don't fragment value specification %s\n", str)
	}
	// Unreached
	return 0
}

func getDoFragValue() int {
	return 0
}

func (f *fragQuiz) initICMP(network, addr string) error {
	var listenNetwork string
	switch network {
	default:
		log.Fatalf("unknown network %v\n", network)
	case "icmp4":
		listenNetwork = "udp4"
	case "icmp6":
		listenNetwork = "udp6"
		// process below
	}

	// Create an unprivileged ICMP listener (yay, no sudo necessary). Only works
	// on MacOS; Linux doesn't allow TTL exceeded messages through the
	// unprivileged ICMP socket.
	c, err := icmp.ListenPacket(listenNetwork, addr)
	if err != nil {
		fmt.Printf("Create %v listener on %s failed: %s\n", network, addr, err)
		return err
	}
	if verbose {
		fmt.Printf("Listening for ICMP on %v\n", c.LocalAddr())
	}
	f.icmp[network] = *c
	return nil
}

func printPMTUD() string {
	sysctlName := "net.inet.tcp.path_mtu_discovery"
	if verbose {
		fmt.Printf("Checking path MTU discovery default in sysctl %v\n", sysctlName)
	}
	pmtud, err := syscall.SysctlUint32(sysctlName)
	if err != nil {
		log.Fatal(err)
	}
	if pmtud != 0 {
		return "on"
	}
	return "off"
}

func printFragSetting(value int) string {
	switch value {
	case 0:
		return "dofrag"
	case 1:
		return "dontfrag"
	}
	log.Fatalf("unknown fragmentation setting %v\n", value)
	// Unreached
	return ""
}

func getFrag(network string, r syscall.RawConn) int {
	var level, opt int
	switch network {
	case "udp4", "tcp4":
		level = syscall.IPPROTO_IP
		opt = unix.IP_DONTFRAG
	case "udp6", "tcp6":
		level = syscall.IPPROTO_IPV6
		opt = unix.IPV6_DONTFRAG
	default:
		log.Panicf("unknown network %v\n", network)
	}

	var err error
	var value int
	rcErr := r.Control(func(fd uintptr) {
		value, err = syscall.GetsockoptInt(int(fd), level, opt)
		if err != nil {
			log.Fatalf("couldn't get don't fragment option for %v socket: %v\n", network, err)
		}
		if verbose {
			fmt.Printf("Got don't fragment bit on %v socket: %v\n", network, value)
		}
	})
	if rcErr != nil {
		log.Fatalf("couldn't execute getsockopt on raw socket for %v socket: %v\n", network, rcErr)
	}
	return value
}

func setFrag(network string, r syscall.RawConn, value int) {
	if verbose {
		fmt.Printf("Setting don't fragment bit on %v socket to %v\n", network, printFragSetting(value))
	}
	var level, opt int
	switch network {
	case "udp4", "tcp4":
		level = syscall.IPPROTO_IP
		opt = unix.IP_DONTFRAG
	case "udp6", "tcp6":
		level = syscall.IPPROTO_IPV6
		opt = unix.IPV6_DONTFRAG
	default:
		log.Panicf("invalid network %v\n", network)
	}
	rcErr := r.Control(func(fd uintptr) {
		err := syscall.SetsockoptInt(int(fd), level, opt, value)
		if err != nil {
			log.Fatalf("couldn't set don't fragment option for %v socket: %v\n", network, err)
		}
	})
	if rcErr != nil {
		log.Fatalf("couldn't execute setsockopt on raw socket for %v socket: %v\n", network, rcErr)
	}
}
