// Copyright (c) Valerie Aurora https://valerieaurora.org/
//
// See COPYING for license information

package main

import (
	"fmt"
	"log"
	"os"
	"strings"
	"syscall"

	"golang.org/x/net/icmp"
)

// Parse the frag values to test string
func parseFrags() {
	frags := strings.Split(fragValuesStr, ",")
	for _, str := range frags {
		var f int
		switch str {
		case "default":
			f = defaultDontFragSetting
		case "dontfrag":
			f = syscall.IP_PMTUDISC_DO
		case "dofrag":
			f = syscall.IP_PMTUDISC_DONT
		case "wantpmtud":
			f = syscall.IP_PMTUDISC_WANT
		case "probe":
			f = syscall.IP_PMTUDISC_PROBE
		default:
			log.Fatalf("Bad don't fragment value specification %s\n", str)
		}
		dontFragValues = append(dontFragValues, f)
	}
}

func getDoFragValue() int {
	return syscall.IP_PMTUDISC_DONT
}

func (f *fragQuiz) initICMP(network, addr string) error {
	var listenNetwork string
	switch network {
	default:
		log.Fatalf("unknown network %v\n", network)
	case "icmp4":
		listenNetwork = "ip4:icmp"
	case "icmp6":
		listenNetwork = "ip6:icmp"
	}

	c, err := icmp.ListenPacket(listenNetwork, addr)
	if err != nil {
		fmt.Printf("Create %v listener on %s failed: %s\n", network, addr, err)
		return err
	}
	if verbose {
		fmt.Printf("Listening for ICMP on %v\n", c.LocalAddr())
	}
	f.icmp[network] = *c
	return nil
}

func printFragSetting(value int) string {
	switch value {
	case syscall.IP_PMTUDISC_DO:
		return "dontfrag (IP_PMTUDISC_DO)"
	case syscall.IP_PMTUDISC_DONT:
		return "dofrag (IP_PMTUDISC_DONT)"
	case syscall.IP_PMTUDISC_WANT:
		return "wantpmtud (IP_PMTUDISC_WANT)"
	case syscall.IP_PMTUDISC_PROBE:
		return "probe (IP_PMTUDISC_PROBE)"
	}
	log.Fatalf("unknown fragmentation setting %v\n", value)
	return ""
}

func printPMTUD() string {
	fileName := "/proc/sys/net/ipv4/ip_no_pmtu_disc"
	if verbose {
		fmt.Printf("Checking path MTU discovery default in %v\n", fileName)
	}
	no_pmtud, err := os.ReadFile(fileName)

	if err != nil {
		log.Fatal(err)
	}
	if string(no_pmtud) != "0" {
		return "on (IP_PMTUDISC_WANT)"
	}
	return "off (IP_PMTUDISC_DONT)"
}

func getFrag(network string, r syscall.RawConn) int {
	var level, opt int
	switch network {
	case "udp4", "tcp4":
		level = syscall.IPPROTO_IP
		opt = syscall.IP_MTU_DISCOVER
	case "udp6", "tcp6":
		level = syscall.IPPROTO_IPV6
		opt = syscall.IPV6_MTU_DISCOVER
	default:
		log.Panicf("unknown network %v\n", network)
	}

	var err error
	var value int
	rcErr := r.Control(func(fd uintptr) {
		value, err = syscall.GetsockoptInt(int(fd), level, opt)
		if err != nil {
			log.Fatalf("couldn't get don't fragment option for %v socket: %v\n", network, err)
		}
		if verbose {
			//fmt.Printf("Got don't fragment bit on %v socket: %v\n", network, value)
		}
	})
	if rcErr != nil {
		log.Fatalf("couldn't execute getsockopt on raw socket for %v socket: %v\n", network, rcErr)
	}
	return value
}

func setFrag(network string, r syscall.RawConn, value int) {
	if verbose {
		fmt.Printf("Setting don't fragment bit on %v socket to %v\n", network, printFragSetting(value))
	}
	var level, opt int
	switch network {
	case "udp4", "tcp4":
		level = syscall.IPPROTO_IP
		opt = syscall.IP_MTU_DISCOVER
	case "udp6", "tcp6":
		level = syscall.IPPROTO_IPV6
		opt = syscall.IPV6_MTU_DISCOVER
	default:
		log.Panicf("invalid network %v\n", network)
	}

	rcErr := r.Control(func(fd uintptr) {
		err := syscall.SetsockoptInt(int(fd), level, opt, value)
		if err != nil {
			log.Fatalf("couldn't set don't fragment option for %v socket: %v\n", network, err)
		}
	})
	if rcErr != nil {
		log.Fatalf("couldn't execute setsockopt on raw socket for %v socket: %v\n", network, rcErr)
	}
}
