// Copyright (c) Valerie Aurora https://valerieaurora.org/

package main

import (
	"errors"
	"fmt"
	"log"
	"net"
)

// Look for some TCP server on the local link.
//
// If it succeeds, it updates the protoAddrPorts list.
func (f *fragQuiz) scanTCPServer(network string) error {
	var allNodesAddr, routersOnlyAddr, icmpNetwork string
	nodesToScan := make(map[string]struct{})
	switch network {
	default:
		log.Fatalf("unknown network %v\n", network)
	case "tcp4":
		log.Printf("%v server scan not implemented yet\n", network)
		return errors.ErrUnsupported
	case "tcp6":
		fmt.Printf("Scanning local link for %v servers\n", network)
		allNodesAddr = "ff02::1%" + defaultInterface
		routersOnlyAddr = "ff02::2%" + defaultInterface
		icmpNetwork = "icmp6"
	}

	_, ok := f.icmp[icmpNetwork]
	if !ok {
		log.Panicf("No %v socket!\n", icmpNetwork)
	}

	// Send an all nodes multicast ping
	f.sendICMP(network, allNodesAddr, 0)
	// Collect the response addresses
	for _, ip, err := f.readICMP(network); err == nil; _, ip, err = f.readICMP(network) {
		if verbose {
			fmt.Printf("Adding %v to node scan list\n", ip)
		}
		nodesToScan[ip] = struct{}{}
	}

	// Send a routers multicast ping
	f.sendICMP(network, routersOnlyAddr, 0)
	// Remove the responses from the nodes
	for _, ip, err := f.readICMP(network); err == nil; _, ip, err = f.readICMP(network) {
		if verbose {
			fmt.Printf("Removing router %v from node scan list\n", ip)
		}
		delete(nodesToScan, ip)
	}

	// Remove all our local interface addresses
	ifs, err := net.Interfaces()
	if err != nil {
		return err
	}
	for _, i := range ifs {
		if i.Name != defaultInterface {
			continue
		}
		addrs, err := i.Addrs()
		if err != nil {
			return err
		}
		for _, a := range addrs {
			ip, _, _ := net.ParseCIDR(a.String())
			if err != nil {
				return err
			}
			ipStr := ip.String() + "%" + i.Name
			if verbose {
				_, ok := nodesToScan[ipStr]
				if !ok {
					fmt.Printf("Attempted to delete local address %v from scan list but not found\n", ipStr)
				} else {
					fmt.Printf("Deleting local addr %v from scan list\n", ipStr)
				}
			}
			delete(nodesToScan, ipStr)
		}
	}

	if verbose {
		fmt.Printf("All non-router non-local nodes replying to multicast ping\n")
		for n := range nodesToScan {
			fmt.Printf("\t%v\n", n)
		}
	}

	// For each response, attempt to connect to a variety of ports
	ports := []string{"80", "443", "22"}
	for n := range nodesToScan {
		for _, p := range ports {
			ipPort := "[" + n + "]:" + p
			if verbose {
				fmt.Printf("Connecting to %v\n", ipPort)
			}
			c, err := net.DialTimeout(network, ipPort, dialTimeout)
			if err == nil {
				fmt.Printf("Using %v for %v tests\n", ipPort, network)
				protoAddrPorts[network] = ipPort
				c.Close()
				return nil
			}
		}
	}
	if verbose {
		fmt.Printf("Couldn't find any open %v ports on the local link\n", network)
	}
	return errors.New("No open servers")
}
